/***************************************************************************
	file:			lcd.c
	product:	API-003
	descr:		Diese API ermoeglicht eine Ausgabe auf einem alphanumerischen LCD 
						mit einem HD44780, oder einem kompatiblen Displaycontroller. 
						Das LCD soll ueber einen 4-bit Datenbus mit dem Mikrocontroller 
						verbunden werden.
	
	date: 		10.2013 - first release

	version 	1.0
	compiler:	Keil Cx51 V9.05

	mc:				AT89C5131A
	crystal:	12MHz, 24MHz

	author: 	Viktor Schabelski info@approbe.de
	licence:	Creative Commons Lizenz
                http://www.stack2Learn.de/cc.html

**************************************************/

/*-----------------------------------------------------------------------------
	includes
---------------------------------------------------*/
#include "lcd.h"
#include "lcd_config.h"


/*-----------------------------------------------------------------------------
	prototype private functions
---------------------------------------------------*/
// Warteschleife
static void LCD_Delay(unsigned int time);
// Lese Nibble
static unsigned char LCD_ReadNibble(void);
//Schreibe Nibble
static errorCodeLCD_t LCD_WriteNibble(unsigned char nibble);
// Busy Flag status
static bit LCD_BFstate(void);
// Schreibe Daten in Instruction Register
static errorCodeLCD_t LCD_SetInstructionRegister(unsigned char dataByte);
// Schreibe Daten in DDRAM
static void LCD_SetDDRAM(unsigned char dataByte);



/*-----------------------------------------------------------------------------
	global variables
---------------------------------------------------*/
#ifndef LOW
#define LOW 0
#endif

#ifndef HIGH
#define HIGH 1
#endif

// erste Zeile
#define LCD_LINE_1						0x01
// zweite Zeile
#define LCD_LINE_2						0x02
// dritte Zeile
#define LCD_LINE_3						0x03
// vierte Zeile
#define LCD_LINE_4						0x04


/*-----------------------------------------------------------------------------
	api
---------------------------------------------------*/

/*=======================================================================*//**
*	@brief		init lcd
---------------------------------------------------*/
void LCD_Init(void){
	LCD_PORT = 0x00;

	LCD_Delay(20000);
	LCD_WriteNibble(0x03);
	LCD_Delay(8000);
	LCD_WriteNibble(0x03);	
	LCD_Delay(500);
		
	LCD_WriteNibble(0x03);
	LCD_WriteNibble(0x02);

	LCD_SetInstructionRegister(FUNCTION_SET);
	LCD_SetInstructionRegister(CURSOR_DISPLAY_SHIFT);

	LCD_SetInstructionRegister(ENTRY_MODE_SET);
	LCD_SetInstructionRegister(DISPLAY_CONTROL);
	LCD_ClearLCD();
}

/*=======================================================================*//**
*	@brief		LCD Hintergrundbeleuchtung EIN / AUS
*	@param		state
---------------------------------------------------*/
void LCD_BackLight (unsigned char state){
	if(state == 1){
	// Hintergrundbeleuchtung EIN
		LCD_BACK_LIGHT = LOW;
	}else{
	// Hintergrundbeleuchtung Aus
		LCD_BACK_LIGHT = HIGH;
	}
}

/*=======================================================================*//**
*	@brief		warte
*	@param		time
---------------------------------------------------*/
static void LCD_Delay(unsigned int time){
	while(time--);
}


/*=======================================================================*//**
*	@brief		Lese 1 nibble (4-bit)
*	@return		nibble: 4-bit data
---------------------------------------------------*/
static unsigned char LCD_ReadNibble(void){
	unsigned char nibble = 0x00;
	LCD_PORT = 0x0F;
	LCD_RW = HIGH;	
	LCD_EN = HIGH;
	LCD_Delay(2);
	nibble = LCD_PORT & 0x0F;
	LCD_Delay(2);
	LCD_EN = LOW;
	return (nibble);
}

/*=======================================================================*//**
*	@brief		Busy Flag Status
*	@return		BF: Busy Flag Status: 1 - busy; 0 - not busy
---------------------------------------------------*/
static bit LCD_BFstate(void){
	unsigned char nibble = 0xFF;
	LCD_RS = LOW;

	nibble = LCD_ReadNibble();	
	LCD_ReadNibble();
	if(nibble & 0x08){
		return HIGH;
	}else{
		return LOW;
	}
}

/*=======================================================================*//**
*	@brief		Schreibe 1 nibble (4-bit)
*	@param		nibble: 4-bit data
*	@return		LCD_TRANSMISSION_FAIL
*	@return		LCD_SUCCESS
---------------------------------------------------*/
static errorCodeLCD_t LCD_WriteNibble(unsigned char nibble){
	unsigned char wait = 100;	
 	LCD_PORT = nibble;
	LCD_RW = LOW;	
	LCD_EN = HIGH;
 	LCD_Delay(2);
	LCD_EN = LOW;
	
	// warte bis LCD Controller fertig ist
 	while ((LCD_BFstate() == HIGH) && (--wait >= 0 ));

	// Fehler Auswertung
	if(wait == 0){
		return (LCD_TRANSMISSION_FAIL);
	}else{
 		return (LCD_SUCCESS);
 	}
}

/*=======================================================================*//**
*	@brief		Schreibe 1 Byte (8-bit) in den Instruction Register
*	@param		dataByte: Byte mit Daten
*	@return		LCD_TRANSMISSION_FAIL
*	@return		LCD_SUCCESS
---------------------------------------------------*/
static errorCodeLCD_t LCD_SetInstructionRegister(unsigned char dataByte){
	errorCodeLCD_t error = LCD_BUSY;
 	LCD_RS = LOW;
 	
	error = LCD_WriteNibble((dataByte >> 4) & 0x0F);
	// Fehlererkennung
	if (error != LCD_SUCCESS){
		return (LCD_TRANSMISSION_FAIL);		
	}

 	error = LCD_WriteNibble(dataByte & 0x0F);
	// Fehlererkennung
	if (error != LCD_SUCCESS){
		return (LCD_TRANSMISSION_FAIL);		
	}else{
		return (LCD_SUCCESS);
	}
}

/*=======================================================================*//**
*	@brief		Schreibe 1 Byte (8-bit) in den DDRAM
*	@param		dataByte: Byte mit Daten
---------------------------------------------------*/
static void LCD_SetDDRAM(unsigned char dataByte){
 	LCD_WriteNibble(((dataByte >> 4) & 0x0F)|0x40); 	// 0x40 => LCD_RS = HIGH;
 	LCD_WriteNibble((dataByte & 0x0F)|0x40);
}

/*=======================================================================*//**
*	@brief		Schreibe String auf den Display
*	@param		line: Zeile, 1 fuer erste Zeile
*	@param		column: Spalte, 0 fuer erste Spalte
*	@param		pString: string
*	@return		LCD_TRANSMISSION_FAIL
*	@return		LCD_SUCCESS
---------------------------------------------------*/
errorCodeLCD_t LCD_PutString (unsigned char line, unsigned char column, unsigned char* pString){
	errorCodeLCD_t error = LCD_BUSY;
	const unsigned char LCD_SET_DDRAM	= 0x80;
	unsigned char postion = LCD_SET_DDRAM + column;

 	if ((column > LCD_COLUMN_SIZE) || (line > LCD_LINE_SIZE)){		/* wenn die Zeile oder Spalte falsch */
 		return (LCD_TRANSMISSION_FAIL);
 	}

	// Adresse berechnen
	switch(line){
		case LCD_LINE_1:
			postion += LCD_ADDRESS_LINE_1;
			error = LCD_SUCCESS;
			break;

		case LCD_LINE_2:
			postion += LCD_ADDRESS_LINE_2;
			error = LCD_SUCCESS;
			break;
			
		case LCD_LINE_3:
			postion += LCD_ADDRESS_LINE_3;
			error = LCD_SUCCESS;
			break;
			
		case LCD_LINE_4:
			postion += LCD_ADDRESS_LINE_4;
			error = LCD_SUCCESS;
			break;

		default:
			return LCD_TRANSMISSION_FAIL;
			break;
	}
	
	// Position vom ersten Zeichen
	LCD_SetInstructionRegister(postion);	

	// String Ausgabe
	while(*pString){
		LCD_SetDDRAM(*pString);
		pString++;
	}
	return (error);
}

/*=======================================================================*//**
*	@brief		Loesche Inhalt des DD-RAM, Setze Adresszaehler auf 0
---------------------------------------------------*/
void LCD_ClearLCD(void){
	// 0x01 steht fuer "clear display" Register
	LCD_SetInstructionRegister (0x01);
}

/*=======================================================================*//**
*	@brief		Setze Adresszaehler auf 0, Setze Cursor auf 0-te Spalte erste Zeile
---------------------------------------------------*/
void LCD_ReturnHome(void){
	// 0x02 steht fuer "return home" Register
	LCD_SetInstructionRegister (0x02);
}



// /*=======================================================================*//**
// *	@brief		test cycle for lcd
// ---------------------------------------------------*/
// void LCD_TestCycle(void){
// 		static unsigned char uch_flag = HIGH;
// 	
// 	if(uch_flag) {
// 		LCD_Init();
// 		uch_flag = LOW;
// 	}

// 	LCD_PutString(1,0,"Hello");
// 	LCD_PutString(2,3,"World");
// }
